#!/usr/bin/python

import os, sys
import codecs
from importer import import_to_sunpinyin_user_dict

def load_google_user_dict (fname):
    result = []
    f = codecs.open (fname, "r", "GB18030")
    for l in f:
        v = l.strip().split()
        utf8str = v[0]

        try:
            freq = int(v[1])
            pystr = '\''.join (v[2:])
        except:
            pystr = '\''.join (v[1:])

        result.append ((pystr, utf8str))

    return result

def main ():
    if len (sys.argv) != 2:
        print "Please specify the Google Pinyin exported user dict file!"
        exit (1)

    google_user_dict = load_google_user_dict(sys.argv[1])
    import_to_sunpinyin_user_dict (google_user_dict)

if __name__ == "__main__":
    main()



#!/usr/bin/python 

# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
# 
# Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
# 
# The contents of this file are subject to the terms of either the GNU Lesser
# General Public License Version 2.1 only ("LGPL") or the Common Development and
# Distribution License ("CDDL")(collectively, the "License"). You may not use this
# file except in compliance with the License. You can obtain a copy of the CDDL at
# http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
# http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
# specific language governing permissions and limitations under the License. When
# distributing the software, include this License Header Notice in each file and
# include the full text of the License in the License file as well as the
# following notice:
# 
# NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
# (CDDL)
# For Covered Software in this distribution, this License shall be governed by the
# laws of the State of California (excluding conflict-of-law provisions).
# Any litigation relating to this License shall be subject to the jurisdiction of
# the Federal Courts of the Northern District of California and the state courts
# of the State of California, with venue lying in Santa Clara County, California.
# 
# Contributor(s):
# 
# If you wish your version of this file to be governed by only the CDDL or only
# the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
# include this software in this distribution under the [CDDL or LGPL Version 2.1]
# license." If you don't indicate a single choice of license, a recipient has the
# option to distribute your version of this file under either the CDDL or the LGPL
# Version 2.1, or to extend the choice of license to its licensees as provided
# above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
# Version 2 license, then the option applies only if the new code is made subject
# to such option by the copyright holder. 

import codecs
from trie import Trie

class IMDict (Trie):
    def __init__(self, fname):
        super (IMDict, self).__init__()
        self.load(fname)

    def load(self, fname):
        file = codecs.open(fname, 'r', 'utf-8')
        for line in file:
            line = line.strip()
            if line[0] not in ('#', '<', '\n'):
                word, id= line.split(' ')[0:2]
                self.add(word, int(id))
        file.close()



#!/usr/bin/python

# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
# 
# Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
# 
# The contents of this file are subject to the terms of either the GNU Lesser
# General Public License Version 2.1 only ("LGPL") or the Common Development and
# Distribution License ("CDDL")(collectively, the "License"). You may not use this
# file except in compliance with the License. You can obtain a copy of the CDDL at
# http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
# http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
# specific language governing permissions and limitations under the License. When
# distributing the software, include this License Header Notice in each file and
# include the full text of the License in the License file as well as the
# following notice:
# 
# NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
# (CDDL)
# For Covered Software in this distribution, this License shall be governed by the
# laws of the State of California (excluding conflict-of-law provisions).
# Any litigation relating to this License shall be subject to the jurisdiction of
# the Federal Courts of the Northern District of California and the state courts
# of the State of California, with venue lying in Santa Clara County, California.
# 
# Contributor(s):
# 
# If you wish your version of this file to be governed by only the CDDL or only
# the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
# include this software in this distribution under the [CDDL or LGPL Version 2.1]
# license." If you don't indicate a single choice of license, a recipient has the
# option to distribute your version of this file under either the CDDL or the LGPL
# Version 2.1, or to extend the choice of license to its licensees as provided
# above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
# Version 2 license, then the option applies only if the new code is made subject
# to such option by the copyright holder. 

__all__ = ['Trie', 'DATrie', 'match_longest', 'get_ambiguious_length']

from math import log
import struct

class Trie (object):
    class TrieNode:
        def __init__ (self):
            self.val = 0
            self.trans = {}

    def __init__(self):
        self.root = Trie.TrieNode()

    def add(self, word, value=1):
        curr_node = self.root
        for ch in word:
            try: 
                curr_node = curr_node.trans[ch]
            except:
                curr_node.trans[ch] = Trie.TrieNode()
                curr_node = curr_node.trans[ch]

        curr_node.val = value

    def walk (self, trienode, ch):
        if ch in trienode.trans:
            trienode = trienode.trans[ch]
            return trienode, trienode.val
        else:
            return None, 0

class FlexibleList (list):
    def __check_size (self, index):
        if index >= len(self):
            self.extend ([0] * (index-len(self)+1))

    def __getitem__ (self, index):
        self.__check_size (index)
        return list.__getitem__(self, index)

    def __setitem__ (self, index, value):
        self.__check_size (index)
        return list.__setitem__(self, index, value)

def character_based_encoder (ch, range=('a', 'z')):
    ret = ord(ch) - ord(range[0]) + 1
    if ret <= 0: ret = ord(range[1]) + 1
    return ret

class DATrie (object):
    def __init__(self, chr_encoder=character_based_encoder):
        self.root = 0
        self.chr_encoder = chr_encoder
        self.clear()

    def clear (self):
        self.base  = FlexibleList ()
        self.check = FlexibleList ()
        self.value = FlexibleList ()

    def walk (self, s, ch):
        c = self.chr_encoder (ch)
        t = abs(self.base[s]) + c

        if t<len(self.check) and self.check[t] == s and self.base[t]:
            if self.value: 
                v = self.value[t]
            else: 
                v = -1 if self.base[t] < 0 else 0
            return t, v
        else:
            return 0, 0

    def find_base (self, s, children, i=1):
        if s == 0 or not children:
            return s

        i = max (i, 1)
        loop_times = 0
        while True:
            for ch in children:
                k = i + self.chr_encoder (ch)
                if self.base[k] or self.check[k] or k == s:
                    loop_times += 1
                    i += int (log (loop_times, 2)) + 1
                    break
            else:
                break

        return i

    def build (self, words, values=None):
        assert (not values or (len(words) == len(values)))
        itval = iter(values) if values else None

        trie = Trie()
        for w in words:
            trie.add (w, itval.next() if itval else -1)

        self.construct_from_trie (trie, values!=None)

    def construct_from_trie (self, trie, with_value=True, progress_cb=None, progress_cb_thr=100):
        nodes = [(trie.root, 0)]
        find_from = 1
        loop_times = 0

        while nodes:
            trienode, s = nodes.pop(0)
            find_from = b = self.find_base (s, trienode.trans, find_from)
            self.base[s] = -b if trienode.val else b
            if with_value: self.value[s] = trienode.val

            for ch in trienode.trans:
                c = self.chr_encoder (ch)
                t = abs(self.base[s]) + c
                self.check[t] = s if s else -1

                nodes.append ((trienode.trans[ch], t))

            loop_times += 1
            if loop_times == progress_cb_thr:
                loop_times = 0
                if progress_cb:
                    progress_cb ()

        for i in xrange (self.chr_encoder (max(trie.root.trans))+1):
            if self.check[i] == -1:
                self.check[i] = 0

    def save (self, fname):
        f = open (fname, 'w+')
        l = len (self.base)

        using_32bits = l > 2**15
        elm_size = 4 if using_32bits else 2
        fmt_str = '%di'%l if using_32bits else '%dh'%l

        # the data types here should be aligned with those in datrie.h
        f.write (struct.pack ('I', l))
        f.write (struct.pack ('H', elm_size))
        f.write (struct.pack ('H', 1 if self.value else 0))

        f.write (struct.pack (fmt_str, *self.base))
        f.write (struct.pack (fmt_str, *self.check))

        if self.value:
            if len(self.value) < l: self.value[l-1] = 0
            f.write (struct.pack ('%di'%l, *self.value))

        f.close()

    def output_static_c_arrays (self, fname):
        f = open(fname, 'w+')
        l = len (self.base)

        type = "int" if l > 2**15 else "short"

        f.write (self.__to_c_array (self.base,  type,  "base"))
        f.write (self.__to_c_array (self.check, type,  "check"))
        f.write (self.__to_c_array (self.value, "int", "value"))

        f.close()

    def __to_c_array (self, array, type, name):
        return "static %s %s[] = {%s};\n\n" % (type, name, ', '.join (str(i) for i in array))

    def load (self, fname):
        f = open (fname, 'r')

        l = struct.unpack ('I', f.read(4))[0]
        elm_size = struct.unpack ('H', f.read(2))[0]
        has_value = struct.unpack ('H', f.read(2))[0]

        fmt_str = '%di'%l if elm_size == 4 else '%dh'%l
        self.base  = struct.unpack (fmt_str, f.read(l*elm_size))
        self.check = struct.unpack (fmt_str, f.read(l*elm_size))
        self.value = struct.unpack ('%di'%l, f.read(l*4)) if has_value else []

        f.close()

def search (trie, word):
    curr_node = trie.root

    for ch in word:
        curr_node, val = trie.walk (curr_node, ch)
        if not curr_node: 
            break
    else:
        return val

    return 0

def match_longest (trie, word):
    l = ret_l = ret_v = 0
    curr_node = trie.root

    for ch in word:
        curr_node, val = trie.walk (curr_node, ch)
        if not curr_node: 
            break

        l += 1
        if val: 
            ret_l, ret_v = l, val

    return ret_v, ret_l

def get_ambiguious_length (trie, str, word_len):
    i = 1
    while i < word_len and i < len(str):
        wid, l = match_longest(trie, str[i:])
        if word_len < i + l:
            word_len = i + l
        i += 1
    return i

def test ():
    from pinyin_data import valid_syllables

    trie = Trie()
    for s in valid_syllables:
        trie.add (s, valid_syllables[s])

    for s in valid_syllables:
        v, l = match_longest (trie, s+'b')
        assert (len(s) == l and valid_syllables[s] == v)

    datrie = DATrie()
    datrie.construct_from_trie (trie)

    datrie.save ('/tmp/trie_test')
    datrie.load ('/tmp/trie_test')

    for s in valid_syllables:
        v, l = match_longest (datrie, s+'b')
        assert (len(s) == l and valid_syllables[s] == v)

    print 'test executed successfully'

if __name__ == "__main__":
    test ()



#!/usr/bin/python 

# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
# 
# Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
# 
# The contents of this file are subject to the terms of either the GNU Lesser
# General Public License Version 2.1 only ("LGPL") or the Common Development and
# Distribution License ("CDDL")(collectively, the "License"). You may not use this
# file except in compliance with the License. You can obtain a copy of the CDDL at
# http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
# http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
# specific language governing permissions and limitations under the License. When
# distributing the software, include this License Header Notice in each file and
# include the full text of the License in the License file as well as the
# following notice:
# 
# NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
# (CDDL)
# For Covered Software in this distribution, this License shall be governed by the
# laws of the State of California (excluding conflict-of-law provisions).
# Any litigation relating to this License shall be subject to the jurisdiction of
# the Federal Courts of the Northern District of California and the state courts
# of the State of California, with venue lying in Santa Clara County, California.
# 
# Contributor(s):
# 
# If you wish your version of this file to be governed by only the CDDL or only
# the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
# include this software in this distribution under the [CDDL or LGPL Version 2.1]
# license." If you don't indicate a single choice of license, a recipient has the
# option to distribute your version of this file under either the CDDL or the LGPL
# Version 2.1, or to extend the choice of license to its licensees as provided
# above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
# Version 2 license, then the option applies only if the new code is made subject
# to such option by the copyright holder. 

initials = ["", "b", "p", "m", "f", "d", "t", "n", "l", "g", "k", "h", "j", "q", "x", "zh", "ch", "sh", "r", "z", "c", "s", "y", "w", ]

finals = ["", "a", "o", "e", "ai", "ei", "ao", "ou", "an", "en", "ang", "eng", "er", "i", "ia", "ie", "iao", "iu", "ian", "in", "iang", "ing", "u", "ua", "uo", "uai", "ui", "uan", "un", "uang", "ong", "v", "ue", "iong", ]

inner_fuzzy_finals = ['ia', 'iao', 'ian', 'iang', 'ie', 'ua', 'uai', 'uan', 'uang', 'ue']

fuzzy_pairs = [
    ('z',       'zh'), 
    ('c',       'ch'), 
    ('s',       'sh'), 
    ('an',      'ang'), 
    ('on',      'ong'), 
    ('en',      'eng'), 
    ('in',      'ing'), 
    ('eng',     'ong'), 
    ('ian',     'iang'), 
    ('uan',     'uang'), 
    ('l',       'n'), 
    ('f',       'h'), 
    ('r',       'l'), 
    ('k',       'g'),
]

auto_correction_pairs = {
    'ign':      'ing',
    'img':      'ing',
    'uei':      'ui',
    'uen':      'un',
    'iou':      'iu',
}

valid_syllables = {
    "a":        0x00010,
    "ai":       0x00040,
    "an":       0x00080,
    "ang":      0x000a0,
    "ao":       0x00060,
    "b":        0x01000,
    "ba":       0x01010,
    "bai":      0x01040,
    "ban":      0x01080,
    "bang":     0x010a0,
    "bao":      0x01060,
    "bei":      0x01050,
    "ben":      0x01090,
    "beng":     0x010b0,
    "bi":       0x010d0,
    "bian":     0x01120,
    "biao":     0x01100,
    "bie":      0x010f0,
    "bin":      0x01130,
    "bing":     0x01150,
    "bo":       0x01020,
    "bu":       0x01160,
    "c":        0x14000,
    "ca":       0x14010,
    "cai":      0x14040,
    "can":      0x14080,
    "cang":     0x140a0,
    "cao":      0x14060,
    "ce":       0x14030,
    "cei":      0x14050,
    "cen":      0x14090,
    "ceng":     0x140b0,
    "ch":       0x10000,
    "cha":      0x10010,
    "chai":     0x10040,
    "chan":     0x10080,
    "chang":    0x100a0,
    "chao":     0x10060,
    "che":      0x10030,
    "chen":     0x10090,
    "cheng":    0x100b0,
    "chi":      0x100d0,
    "chong":    0x101e0,
    "chou":     0x10070,
    "chu":      0x10160,
    "chua":     0x10170,
    "chuai":    0x10190,
    "chuan":    0x101b0,
    "chuang":   0x101d0,
    "chui":     0x101a0,
    "chun":     0x101c0,
    "chuo":     0x10180,
    "ci":       0x140d0,
    "cong":     0x141e0,
    "cou":      0x14070,
    "cu":       0x14160,
    "cuan":     0x141b0,
    "cui":      0x141a0,
    "cun":      0x141c0,
    "cuo":      0x14180,
    "d":        0x05000,
    "da":       0x05010,
    "dai":      0x05040,
    "dan":      0x05080,
    "dang":     0x050a0,
    "dao":      0x05060,
    "de":       0x05030,
    "dei":      0x05050,
    "den":      0x05090,
    "deng":     0x050b0,
    "di":       0x050d0,
    "dia":      0x050e0,
    "dian":     0x05120,
    "diao":     0x05100,
    "die":      0x050f0,
    "ding":     0x05150,
    "diu":      0x05110,
    "dong":     0x051e0,
    "dou":      0x05070,
    "du":       0x05160,
    "duan":     0x051b0,
    "dui":      0x051a0,
    "dun":      0x051c0,
    "duo":      0x05180,
    "e":        0x00030,
    "ei":       0x00050,
    "en":       0x00090,
    "eng":      0x000b0,
    "er":       0x000c0,
    "f":        0x04000,
    "fa":       0x04010,
    "fan":      0x04080,
    "fang":     0x040a0,
    "fei":      0x04050,
    "fen":      0x04090,
    "feng":     0x040b0,
    "fiao":     0x04100,
    "fo":       0x04020,
    "fou":      0x04070,
    "fu":       0x04160,
    "g":        0x09000,
    "ga":       0x09010,
    "gai":      0x09040,
    "gan":      0x09080,
    "gang":     0x090a0,
    "gao":      0x09060,
    "ge":       0x09030,
    "gei":      0x09050,
    "gen":      0x09090,
    "geng":     0x090b0,
    "gong":     0x091e0,
    "gou":      0x09070,
    "gu":       0x09160,
    "gua":      0x09170,
    "guai":     0x09190,
    "guan":     0x091b0,
    "guang":    0x091d0,
    "gui":      0x091a0,
    "gun":      0x091c0,
    "guo":      0x09180,
    "h":        0x0b000,
    "ha":       0x0b010,
    "hai":      0x0b040,
    "han":      0x0b080,
    "hang":     0x0b0a0,
    "hao":      0x0b060,
    "he":       0x0b030,
    "hei":      0x0b050,
    "hen":      0x0b090,
    "heng":     0x0b0b0,
    "hong":     0x0b1e0,
    "hou":      0x0b070,
    "hu":       0x0b160,
    "hua":      0x0b170,
    "huai":     0x0b190,
    "huan":     0x0b1b0,
    "huang":    0x0b1d0,
    "hui":      0x0b1a0,
    "hun":      0x0b1c0,
    "huo":      0x0b180,
    "j":        0x0c000,
    "ji":       0x0c0d0,
    "jia":      0x0c0e0,
    "jian":     0x0c120,
    "jiang":    0x0c140,
    "jiao":     0x0c100,
    "jie":      0x0c0f0,
    "jin":      0x0c130,
    "jing":     0x0c150,
    "jiong":    0x0c210,
    "jiu":      0x0c110,
    "ju":       0x0c160,
    "juan":     0x0c1b0,
    "jue":      0x0c200,
    "jun":      0x0c1c0,
    "k":        0x0a000,
    "ka":       0x0a010,
    "kai":      0x0a040,
    "kan":      0x0a080,
    "kang":     0x0a0a0,
    "kao":      0x0a060,
    "ke":       0x0a030,
    "kei":      0x0a050,
    "ken":      0x0a090,
    "keng":     0x0a0b0,
    "kong":     0x0a1e0,
    "kou":      0x0a070,
    "ku":       0x0a160,
    "kua":      0x0a170,
    "kuai":     0x0a190,
    "kuan":     0x0a1b0,
    "kuang":    0x0a1d0,
    "kui":      0x0a1a0,
    "kun":      0x0a1c0,
    "kuo":      0x0a180,
    "l":        0x08000,
    "la":       0x08010,
    "lai":      0x08040,
    "lan":      0x08080,
    "lang":     0x080a0,
    "lao":      0x08060,
    "le":       0x08030,
    "lei":      0x08050,
    "leng":     0x080b0,
    "li":       0x080d0,
    "lia":      0x080e0,
    "lian":     0x08120,
    "liang":    0x08140,
    "liao":     0x08100,
    "lie":      0x080f0,
    "lin":      0x08130,
    "ling":     0x08150,
    "liu":      0x08110,
    "lo":       0x08020,
    "long":     0x081e0,
    "lou":      0x08070,
    "lu":       0x08160,
    "luan":     0x081b0,
    "lue":      0x08200,
    "lun":      0x081c0,
    "luo":      0x08180,
    "lv":       0x081f0,
    "m":        0x03000,
    "ma":       0x03010,
    "mai":      0x03040,
    "man":      0x03080,
    "mang":     0x030a0,
    "mao":      0x03060,
    "me":       0x03030,
    "mei":      0x03050,
    "men":      0x03090,
    "meng":     0x030b0,
    "mi":       0x030d0,
    "mian":     0x03120,
    "miao":     0x03100,
    "mie":      0x030f0,
    "min":      0x03130,
    "ming":     0x03150,
    "miu":      0x03110,
    "mo":       0x03020,
    "mou":      0x03070,
    "mu":       0x03160,
    "n":        0x07000,
    "na":       0x07010,
    "nai":      0x07040,
    "nan":      0x07080,
    "nang":     0x070a0,
    "nao":      0x07060,
    "ne":       0x07030,
    "nei":      0x07050,
    "nen":      0x07090,
    "neng":     0x070b0,
    "ni":       0x070d0,
    "nian":     0x07120,
    "niang":    0x07140,
    "niao":     0x07100,
    "nie":      0x070f0,
    "nin":      0x07130,
    "ning":     0x07150,
    "niu":      0x07110,
    "nong":     0x071e0,
    "nou":      0x07070,
    "nu":       0x07160,
    "nuan":     0x071b0,
    "nue":      0x07200,
    "nun":      0x071c0,
    "nuo":      0x07180,
    "nv":       0x071f0,
    "o":        0x00020,
    "ou":       0x00070,
    "p":        0x02000,
    "pa":       0x02010,
    "pai":      0x02040,
    "pan":      0x02080,
    "pang":     0x020a0,
    "pao":      0x02060,
    "pei":      0x02050,
    "pen":      0x02090,
    "peng":     0x020b0,
    "pi":       0x020d0,
    "pian":     0x02120,
    "piao":     0x02100,
    "pie":      0x020f0,
    "pin":      0x02130,
    "ping":     0x02150,
    "po":       0x02020,
    "pou":      0x02070,
    "pu":       0x02160,
    "q":        0x0d000,
    "qi":       0x0d0d0,
    "qia":      0x0d0e0,
    "qian":     0x0d120,
    "qiang":    0x0d140,
    "qiao":     0x0d100,
    "qie":      0x0d0f0,
    "qin":      0x0d130,
    "qing":     0x0d150,
    "qiong":    0x0d210,
    "qiu":      0x0d110,
    "qu":       0x0d160,
    "quan":     0x0d1b0,
    "que":      0x0d200,
    "qun":      0x0d1c0,
    "r":        0x12000,
    "ran":      0x12080,
    "rang":     0x120a0,
    "rao":      0x12060,
    "re":       0x12030,
    "ren":      0x12090,
    "reng":     0x120b0,
    "ri":       0x120d0,
    "rong":     0x121e0,
    "rou":      0x12070,
    "ru":       0x12160,
    "ruan":     0x121b0,
    "rui":      0x121a0,
    "run":      0x121c0,
    "ruo":      0x12180,
    "s":        0x15000,
    "sa":       0x15010,
    "sai":      0x15040,
    "san":      0x15080,
    "sang":     0x150a0,
    "sao":      0x15060,
    "se":       0x15030,
    "sen":      0x15090,
    "seng":     0x150b0,
    "sh":       0x11000,
    "sha":      0x11010,
    "shai":     0x11040,
    "shan":     0x11080,
    "shang":    0x110a0,
    "shao":     0x11060,
    "she":      0x11030,
    "shei":     0x11050,
    "shen":     0x11090,
    "sheng":    0x110b0,
    "shi":      0x110d0,
    "shou":     0x11070,
    "shu":      0x11160,
    "shua":     0x11170,
    "shuai":    0x11190,
    "shuan":    0x111b0,
    "shuang":   0x111d0,
    "shui":     0x111a0,
    "shun":     0x111c0,
    "shuo":     0x11180,
    "si":       0x150d0,
    "song":     0x151e0,
    "sou":      0x15070,
    "su":       0x15160,
    "suan":     0x151b0,
    "sui":      0x151a0,
    "sun":      0x151c0,
    "suo":      0x15180,
    "t":        0x06000,
    "ta":       0x06010,
    "tai":      0x06040,
    "tan":      0x06080,
    "tang":     0x060a0,
    "tao":      0x06060,
    "te":       0x06030,
    "tei":      0x06050,
    "teng":     0x060b0,
    "ti":       0x060d0,
    "tian":     0x06120,
    "tiao":     0x06100,
    "tie":      0x060f0,
    "ting":     0x06150,
    "tong":     0x061e0,
    "tou":      0x06070,
    "tu":       0x06160,
    "tuan":     0x061b0,
    "tui":      0x061a0,
    "tun":      0x061c0,
    "tuo":      0x06180,
    "w":        0x17000,
    "wa":       0x17010,
    "wai":      0x17040,
    "wan":      0x17080,
    "wang":     0x170a0,
    "wei":      0x17050,
    "wen":      0x17090,
    "weng":     0x170b0,
    "wo":       0x17020,
    "wu":       0x17160,
    "x":        0x0e000,
    "xi":       0x0e0d0,
    "xia":      0x0e0e0,
    "xian":     0x0e120,
    "xiang":    0x0e140,
    "xiao":     0x0e100,
    "xie":      0x0e0f0,
    "xin":      0x0e130,
    "xing":     0x0e150,
    "xiong":    0x0e210,
    "xiu":      0x0e110,
    "xu":       0x0e160,
    "xuan":     0x0e1b0,
    "xue":      0x0e200,
    "xun":      0x0e1c0,
    "y":        0x16000,
    "ya":       0x16010,
    "yai":      0x16040,
    "yan":      0x16080,
    "yang":     0x160a0,
    "yao":      0x16060,
    "ye":       0x16030,
    "yi":       0x160d0,
    "yin":      0x16130,
    "ying":     0x16150,
    "yo":       0x16020,
    "yong":     0x161e0,
    "you":      0x16070,
    "yu":       0x16160,
    "yuan":     0x161b0,
    "yue":      0x16200,
    "yun":      0x161c0,
    "z":        0x13000,
    "za":       0x13010,
    "zai":      0x13040,
    "zan":      0x13080,
    "zang":     0x130a0,
    "zao":      0x13060,
    "ze":       0x13030,
    "zei":      0x13050,
    "zen":      0x13090,
    "zeng":     0x130b0,
    "zh":       0x0f000,
    "zha":      0x0f010,
    "zhai":     0x0f040,
    "zhan":     0x0f080,
    "zhang":    0x0f0a0,
    "zhao":     0x0f060,
    "zhe":      0x0f030,
    "zhei":     0x0f050,
    "zhen":     0x0f090,
    "zheng":    0x0f0b0,
    "zhi":      0x0f0d0,
    "zhong":    0x0f1e0,
    "zhou":     0x0f070,
    "zhu":      0x0f160,
    "zhua":     0x0f170,
    "zhuai":    0x0f190,
    "zhuan":    0x0f1b0,
    "zhuang":   0x0f1d0,
    "zhui":     0x0f1a0,
    "zhun":     0x0f1c0,
    "zhuo":     0x0f180,
    "zi":       0x130d0,
    "zong":     0x131e0,
    "zou":      0x13070,
    "zu":       0x13160,
    "zuan":     0x131b0,
    "zui":      0x131a0,
    "zun":      0x131c0,
    "zuo":      0x13180,
}

def decode_syllable (s):
    return initials[(s>>12)], finals[(s&0x00ff0)>>4]

def init_fuzzy_map (fuzzy_pairs):
    fuzzy_map = {}
    for i, j in fuzzy_pairs:
        fuzzy_map.setdefault (i, []).append (j)
        fuzzy_map.setdefault (j, []).append (i)

    return fuzzy_map

fuzzy_map = init_fuzzy_map (fuzzy_pairs)

def get_fuzzy_syllables (syllable):
    i, f = decode_syllable (syllable)
    iset = fuzzy_map.setdefault(i, []) + [i]
    fset = fuzzy_map.setdefault(f, []) + [f]
    sset = [valid_syllables[i+f] for i in iset for f in fset if i+f in valid_syllables]
    sset.remove (syllable)
    return sset

def gen_suffix_trie (fname):
    from trie import Trie, DATrie
    
    trie = Trie ()
    pytrie = DATrie ()

    for s in valid_syllables:
        trie.add (s[::-1], valid_syllables[s])
    
    pytrie.construct_from_trie (trie)
    pytrie.output_static_c_arrays (fname)

def gen_fuzzy_syllable_pairs_tables ():
    fuzzy_pro_syllables = [s for s in valid_syllables if s[1:] in valid_syllables and s[0] in initials and s not in initials]
    fuzzy_pre_syllables = [s for s in valid_syllables if s[:-1] in valid_syllables and s[-1] in initials and s not in initials]

    initial_sets = set([s[0] for s in fuzzy_pro_syllables]) & set([s[-1] for s in fuzzy_pre_syllables])

    fuzzy_pro_syllables  = [s for s in fuzzy_pro_syllables if s[0] in initial_sets]
    fuzzy_pre_syllables  = [s for s in fuzzy_pre_syllables if s[-1] in initial_sets]

    print "static const unsigned fuzzy_pre_syllables [] = {"
    for s in fuzzy_pre_syllables:
        print "    %-12s %-12s %-12s /* %s */" % ("0x%05x," % valid_syllables[s[:-1]], "'%s'," % s[-1], "0x%05x," % valid_syllables[s], s)
    print "    0x0,"
    print "};\n"

    print "static const unsigned fuzzy_pro_syllables [] = {"
    for s in fuzzy_pro_syllables:
        print "    %-12s %-12s %-12s /* %s */" % ("0x%05x," % valid_syllables[s], "'%s'," % s[0], "0x%05x," % valid_syllables[s[1:]], s)
    print "    0x0,"
    print "};\n"

def gen_inner_fuzzy_syllable_tables ():
    print "static const unsigned fuzzy_finals_map[] = {"
    for s in inner_fuzzy_finals:
        print "    %-12s %-12s %-12s /* %-4s -> %-4s len %d */" % ("0x%02x," % finals.index(s), "0x%02x," % valid_syllables[s[1:]], "%d," % (len(s)-1,),  s, s[1:], len(s)-1)
    print "};\n"

if __name__ == "__main__":
    gen_suffix_trie ("../src/pinyin/quanpin_trie.h")
    gen_inner_fuzzy_syllable_tables ()
    gen_fuzzy_syllable_pairs_tables ()



#!/usr/bin/python

# thanks for the reverse engineering efforts of following projects/peoples:
# http://code.google.com/p/imewlconverter
# http://code.google.com/p/ibus-cloud-pinyin
# http://forum.ubuntu.org.cn/viewtopic.php?f=8&t=250136&start=0

from importer import import_to_sunpinyin_user_dict
import struct
import os, sys

def read_utf16_str (f, offset=-1, len=2):
    if offset >= 0:
        f.seek(offset)
    str = f.read(len)
    return str.decode('UTF-16LE')

def read_uint16 (f):
    return struct.unpack ('<H', f.read(2))[0]

def get_word_from_sogou_cell_dict (fname):
    f = open (fname, 'rb')
    file_size = os.path.getsize (fname)
    
    hz_offset = 0
    mask = struct.unpack ('B', f.read(128)[4])[0]
    if mask == 0x44:
        hz_offset = 0x2628
    elif mask == 0x45:
        hz_offset = 0x26c4
    else:
        sys.exit(1)
    
    title   = read_utf16_str (f, 0x130, 0x338  - 0x130)
    type    = read_utf16_str (f, 0x338, 0x540  - 0x338)
    desc    = read_utf16_str (f, 0x540, 0xd40  - 0x540)
    samples = read_utf16_str (f, 0xd40, 0x1540 - 0xd40)
    
    py_map = {}
    f.seek(0x1540+4)
    
    while 1:
        py_code = read_uint16 (f)
        py_len  = read_uint16 (f)
        py_str  = read_utf16_str (f, -1, py_len)
    
        if py_code not in py_map:
            py_map[py_code] = py_str
    
        if py_str == 'zuo':
            break
    
    f.seek(hz_offset)
    while f.tell() != file_size:
        word_count   = read_uint16 (f)
        pinyin_count = read_uint16 (f) / 2
    
        py_set = []
        for i in range(pinyin_count):
            py_id = read_uint16(f)
            py_set.append(py_map[py_id])
        py_str = "'".join (py_set)
    
        for i in range(word_count):
            word_len = read_uint16(f)
            word_str = read_utf16_str (f, -1, word_len)
            f.read(12)  # simply ignore word frequence info
            yield py_str, word_str

    f.close()

def main ():
    if len (sys.argv) != 2:
        print "Please specify the Sogou PinYin Cell dict file!"
        exit (1)

    generator = get_word_from_sogou_cell_dict (sys.argv[1])
    import_to_sunpinyin_user_dict (generator)

if __name__ == "__main__":
    main()



# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
# 
# Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
# 
# The contents of this file are subject to the terms of either the GNU Lesser
# General Public License Version 2.1 only ("LGPL") or the Common Development and
# Distribution License ("CDDL")(collectively, the "License"). You may not use this
# file except in compliance with the License. You can obtain a copy of the CDDL at
# http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
# http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
# specific language governing permissions and limitations under the License. When
# distributing the software, include this License Header Notice in each file and
# include the full text of the License in the License file as well as the
# following notice:
# 
# NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
# (CDDL)
# For Covered Software in this distribution, this License shall be governed by the
# laws of the State of California (excluding conflict-of-law provisions).
# Any litigation relating to this License shall be subject to the jurisdiction of
# the Federal Courts of the Northern District of California and the state courts
# of the State of California, with venue lying in Santa Clara County, California.
# 
# Contributor(s):
# 
# If you wish your version of this file to be governed by only the CDDL or only
# the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
# include this software in this distribution under the [CDDL or LGPL Version 2.1]
# license." If you don't indicate a single choice of license, a recipient has the
# option to distribute your version of this file under either the CDDL or the LGPL
# Version 2.1, or to extend the choice of license to its licensees as provided
# above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
# Version 2 license, then the option applies only if the new code is made subject
# to such option by the copyright holder. 

cdef extern from "slm.h":

    ctypedef union CThreadSlm_TState "CThreadSlm::TState":
        void setIdx (unsigned int idx)
        void setLevel (unsigned int level)
        unsigned int getLevel()
        unsigned int getIdx()

    ctypedef struct CThreadSlm "CThreadSlm":
        bint load(char *filename, bint MMap)
        bint isUseLogPr()
        void free()
        double transferNegLog (CThreadSlm_TState history, unsigned int wid, CThreadSlm_TState result)
        double transfer (CThreadSlm_TState history, unsigned int wid, CThreadSlm_TState result)
        CThreadSlm_TState history_state_of(CThreadSlm_TState st)
        CThreadSlm_TState historify(CThreadSlm_TState st)
        unsigned int lastWordId(CThreadSlm_TState st)

    CThreadSlm *new_CThreadSlm "new CThreadSlm" ()
    void del_CThreadSlm "delete" (CThreadSlm *slm)

cdef class SlmState:
    cdef public int level, idx
    def __cinit__(self, level=0, idx=0):
        self.level = level
        self.idx = idx

    def __richcmp__ (self, other, op):
        equal = self.level == other.level and self.idx == other.idx
        if   op == 2:   return equal
        elif op == 3:   return not equal
        else:           return NotImplemented

    def __hash__ (self):
        return ((self.level, self.idx)).__hash__()

    def __str__(self):
        return "[level=%d, idx=%d]" % (self.level, self.idx)

cdef inline pystate_to_cstate (state, CThreadSlm_TState *st):
    st.setLevel (state.level)
    st.setIdx (state.idx)

cdef class Slm:
    cdef CThreadSlm *thisptr

    def __cinit__(self):
        self.thisptr = new_CThreadSlm()

    def __dealloc__(self):
        del_CThreadSlm (self.thisptr)

    def load (self, fname):
        return self.thisptr.load(fname, True)

    def is_using_log_pr(self):
        return self.thisptr.isUseLogPr()

    def free(self):
        self.thisptr.free()

    def transfer_neglog(self, history, wid):
        cdef CThreadSlm_TState his, ret
        pystate_to_cstate (history, &his)
        pr = self.thisptr.transferNegLog(his, wid, ret)
        return pr, SlmState(ret.getLevel(), ret.getIdx())

    def transfer(self, history, wid):
        cdef CThreadSlm_TState his, ret
        pystate_to_cstate (history, &his)
        pr = self.thisptr.transfer(his, wid, ret)
        return pr, SlmState(ret.getLevel(), ret.getIdx())
    
    def history_state_of(self, state):
        cdef CThreadSlm_TState st, ret
        pystate_to_cstate (state, &st)
        ret = self.thisptr.history_state_of(st)
        return SlmState(ret.getLevel(), ret.getIdx())

    def historify (self, state):
        cdef CThreadSlm_TState st
        pystate_to_cstate (state, &st)
        self.thisptr.historify(st)
        state.level = int(st.getLevel())
        state.idx = int(st.getIdx())

    def last_word_id (self, state):
        cdef CThreadSlm_TState st
        pystate_to_cstate (state, &st)
        return self.thisptr.lastWordId(st)



# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
# 
# Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
# 
# The contents of this file are subject to the terms of either the GNU Lesser
# General Public License Version 2.1 only ("LGPL") or the Common Development and
# Distribution License ("CDDL")(collectively, the "License"). You may not use this
# file except in compliance with the License. You can obtain a copy of the CDDL at
# http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
# http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
# specific language governing permissions and limitations under the License. When
# distributing the software, include this License Header Notice in each file and
# include the full text of the License in the License file as well as the
# following notice:
# 
# NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
# (CDDL)
# For Covered Software in this distribution, this License shall be governed by the
# laws of the State of California (excluding conflict-of-law provisions).
# Any litigation relating to this License shall be subject to the jurisdiction of
# the Federal Courts of the Northern District of California and the state courts
# of the State of California, with venue lying in Santa Clara County, California.
# 
# Contributor(s):
# 
# If you wish your version of this file to be governed by only the CDDL or only
# the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
# include this software in this distribution under the [CDDL or LGPL Version 2.1]
# license." If you don't indicate a single choice of license, a recipient has the
# option to distribute your version of this file under either the CDDL or the LGPL
# Version 2.1, or to extend the choice of license to its licensees as provided
# above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
# Version 2 license, then the option applies only if the new code is made subject
# to such option by the copyright holder. 

cdef extern from "Python.h":
    ctypedef struct PyUnicodeObject:
        pass
    ctypedef unsigned wchar_t
    ctypedef wchar_t * const_wchar_t_ptr "const wchar_t *"
    object PyUnicode_FromWideChar (wchar_t *, Py_ssize_t)
    Py_ssize_t PyUnicode_AsWideChar (PyUnicodeObject *, wchar_t *, Py_ssize_t)

cdef extern from "portability.h":
    ctypedef unsigned TWCHAR
    ctypedef TWCHAR * const_TWCHAR_ptr "const TWCHAR *"
    unsigned WCSLEN (const_TWCHAR_ptr ws)

cdef extern from "pytrie.h":
    ctypedef struct CPinyinTrie_TWord "CPinyinTrie::TWordIdInfo":
        bint     m_bSeen
        unsigned m_cost
        unsigned m_csLevel
        unsigned m_id
        unsigned m_len

    ctypedef struct CPinyinTrie_TNode "CPinyinTrie::TNode":
        unsigned m_nWordId
        CPinyinTrie_TWord * getWordIdPtr ()

    ctypedef struct CPinyinTrie "CPinyinTrie":
        bint load(char *filename)
        void free()
        int getWordCount ()
        CPinyinTrie_TNode * getRootNode ()
        CPinyinTrie_TNode * transfer (CPinyinTrie_TNode *, unsigned)
        const_TWCHAR_ptr getitem "operator []" (unsigned)
        int getSymbolId (const_TWCHAR_ptr)
        bint isValid (CPinyinTrie_TNode*, bint, unsigned)

    CPinyinTrie *new_CPinyinTrie "new CPinyinTrie" ()
    void del_CPinyinTrie "delete" (CPinyinTrie *pytrie)

cdef class WordInfo:
    cdef public bint seen
    cdef public int wid, cost, length, charset_level

    def __cinit__ (self, wid, seen=True, cost=0, length=0, charset_level=0):
        self.wid = wid
        self.seen = seen
        self.cost = cost
        self.length = length
        self.charset_level = charset_level

    def __str__ (self):
        return "wid=%d, seen=%d, cost=%d, length=%d, charset_level=%d" % \
               (self.wid, self.seen, self.cost, self.length, self.charset_level)

cdef class PinyinTrieNode:
    cdef CPinyinTrie_TNode *pnode

    def get_words (self):
        words = []
        cdef CPinyinTrie_TWord *p= <CPinyinTrie_TWord*> self.pnode.getWordIdPtr ()
        for i in xrange (self.pnode.m_nWordId):
            words.append (WordInfo(p[i].m_id, p[i].m_bSeen, p[i].m_cost, p[i].m_len, p[i].m_csLevel))
        return words 

cdef class PinyinTrie:
    cdef CPinyinTrie *thisptr

    def __cinit__ (self):
        self.thisptr = new_CPinyinTrie ()

    def __dealloc__ (self):
        del_CPinyinTrie (self.thisptr)

    def load (self, fname):
        return self.thisptr.load (fname)

    def free (self):
        self.thisptr.free ()

    def get_word_count (self):
        return self.thisptr.getWordCount()

    def get_root_node (self):
        cdef CPinyinTrie_TNode * pnode = <CPinyinTrie_TNode*> self.thisptr.getRootNode ()
        node = PinyinTrieNode ()
        (<PinyinTrieNode>node).pnode = pnode
        return node

    def transfer (self, node, syllable):
        cdef CPinyinTrie_TNode *pnode = (<PinyinTrieNode>node).pnode
        pnode = <CPinyinTrie_TNode*> self.thisptr.transfer (pnode, <int>syllable)
        if not pnode:
            return None

        node = PinyinTrieNode ()
        (<PinyinTrieNode>node).pnode = pnode
        return node

    def __getitem__ (self, idx):
        if idx<0 or idx>=self.thisptr.getWordCount(): return ''
        cdef const_TWCHAR_ptr cwstr = self.thisptr.getitem (idx)
        return PyUnicode_FromWideChar (<const_wchar_t_ptr>cwstr, WCSLEN(cwstr))

    def get_symbol_id (self, symbol):
        cdef wchar_t buf[2]
        if len (symbol) != 1: return 0
        PyUnicode_AsWideChar (<PyUnicodeObject*> symbol, buf, sizeof(buf))
        return self.thisptr.getSymbolId (<const_TWCHAR_ptr> buf)

    def is_valid (self, node, allowNonComplete=True, csLevel=0):
        cdef CPinyinTrie_TNode *pnode = (<PinyinTrieNode>node).pnode
        return self.thisptr.isValid (pnode, allowNonComplete, csLevel)



#!/usr/bin/python 
# -*- coding: UTF-8 -*-

# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
# 
# Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
# 
# The contents of this file are subject to the terms of either the GNU Lesser
# General Public License Version 2.1 only ("LGPL") or the Common Development and
# Distribution License ("CDDL")(collectively, the "License"). You may not use this
# file except in compliance with the License. You can obtain a copy of the CDDL at
# http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
# http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
# specific language governing permissions and limitations under the License. When
# distributing the software, include this License Header Notice in each file and
# include the full text of the License in the License file as well as the
# following notice:
# 
# NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
# (CDDL)
# For Covered Software in this distribution, this License shall be governed by the
# laws of the State of California (excluding conflict-of-law provisions).
# Any litigation relating to this License shall be subject to the jurisdiction of
# the Federal Courts of the Northern District of California and the state courts
# of the State of California, with venue lying in Santa Clara County, California.
# 
# Contributor(s):
# 
# If you wish your version of this file to be governed by only the CDDL or only
# the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
# include this software in this distribution under the [CDDL or LGPL Version 2.1]
# license." If you don't indicate a single choice of license, a recipient has the
# option to distribute your version of this file under either the CDDL or the LGPL
# Version 2.1, or to extend the choice of license to its licensees as provided
# above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
# Version 2 license, then the option applies only if the new code is made subject
# to such option by the copyright holder. 

import os
import mmap
import struct
import heapq
import tempfile

class NGram:
    key = ()
    freq = 0

    def __init__(self, key, freq):
        self.key = key
        self.freq = freq

    def __cmp__(self,other):
        return cmp(self.key, other.key)

    def __str__(self):
        return "ngram: " + self.key.__str__() + " freq: " + str(self.freq)

def read_ch_sentences(file):
    nesting = 0
    buf = ""
    for line in file:
        if buf and (line[0].isspace() or len(buf) <= 40):
            yield buf
            buf, nesting = "", 0

        for ch in line:
            if ch.isspace():
                continue

            if ch in u"（“《":
		nesting +=1
            elif ch in u"）”》":
		nesting -=1

            if ch in u"；。！？…" and nesting == 0:
                if buf:
                    yield buf + ch
                    buf, nesting = "", 0
            else:
                buf += ch
    if buf:
        yield buf

def mergesort (iters):
        heap=[]

        for it in iters:
            try:
                heap.append((it.next(), it))
            except StopIteration:
                pass

        heapq.heapify(heap)

        while heap:
            val, it = heap[0]
            yield val

            try:
                heapq.heapreplace(heap, (it.next(),it))
            except StopIteration:
                heapq.heappop(heap)

def read_ngrams (fname, n):
    file = open(fname, "r")
    fsize = os.path.getsize(fname)
    mem = mmap.mmap(file.fileno(), fsize, mmap.MAP_SHARED, mmap.PROT_READ)

    while True:
        ngram = mem.read((n+1)*4)
        if ngram:
            data = struct.unpack('%dl' % (n+1), ngram)
            yield NGram(data[:n], data[n])
        else:
            break

    mem.close()
    file.close()

class MMArray:
    __file = __mem = None
    __realsize = __capsize = 0

    def __init__(self, elmsize=1, fname=None, capsize=1024*1024):
        self.__elmsize = elmsize

        if not fname:
            fno, self.__fname = tempfile.mkstemp("-mmarray", "pyslm-")
            self.__file = os.fdopen (fno, "w+")
            self.__enlarge(capsize)
        else:
            self.fromfile(fname)

    def fromfile(self, fname):
        if not os.path.exists(fname):
            raise "The file '%s' does not exist!"

        fsize = os.path.getsize(fname)
        if fsize == 0:
            raise "The size of file '%s' is zero!" % fname

        if self.__mem: self.__mem.close()
        if self.__file: self.__file.close()

        self.__file = open (fname, "r+")
        self.__mem = mmap.mmap(self.__file.fileno(), fsize)
        self.__realsize = self.__capsize = fsize/self.__elmsize

    def tofile(self, fname):
        if fname == self.__file.name:
            raise "Can not dump the array to currently mapping file!"
        tf = open(fname, "w+")
        bsize = self.__realsize * self.__elmsize
        tf.write (self.__mem[:bsize])
        tf.close()

    def __enlarge(self, capsize):
        if self.__capsize >= capsize:
            return
        
        self.__capsize = capsize
        self.__file.seek(self.__elmsize * self.__capsize - 1)
        self.__file.write('\0')
        self.__file.flush()

        if (self.__mem): self.__mem.close()
        self.__mem = mmap.mmap(self.__file.fileno(), self.__file.tell())

    def __del__ (self):
        bsize = self.__realsize * self.__elmsize
        self.__file.truncate (bsize)
        self.__file.close()
        if self.__mem: self.__mem.close()
        os.remove(self.__fname)

    def __getitem__(self, idx):
        if idx < -self.__realsize or idx >= self.__realsize:
            raise IndexError
        return self.__access(idx)

    def __setitem__(self, idx, buf):
        if idx < -self.__realsize or idx >= self.__realsize:
            raise IndexError
        if type(buf) != type("") or len(buf) != self.__elmsize:
            raise "Not a string, or the buffer size is incorrect!"
        self.__access(idx, buf)

    def __access (self, idx, buf=None):
        if idx < 0: idx = self.__realsize + idx
        start = idx * self.__elmsize
        end = start + self.__elmsize
        if not buf: return self.__mem[start:end]
        self.__mem[start:end] = buf

    def size(self):
        return self.__realsize

    def append(self, buf):
        if type(buf) != type("") or len(buf) != self.__elmsize:
            raise "Not a string, or the buffer size is incorrect!"

        if self.__realsize >= self.__capsize:
            self.__enlarge(self.__capsize*2)

        self.__access(self.__realsize, buf)
        self.__realsize += 1

    def __iter__(self):
        for i in xrange(0, self.__realsize):
            yield self.__access(i)

    def truncate(self, tsize):
        if self.__realsize >= tsize:
            self.__realsize = tsize



#!/usr/bin/python

import os
import sqlite3 as sqlite
from importer import import_to_sunpinyin_user_dict

def load_fit_user_dict ():
    homedir = os.environ.get("HOME")
    db = sqlite.connect (homedir+"/Library/FunInputToy/py.sqlite")
    sqlstring = "SELECT code, string FROM words WHERE is_new==1"
    result = list (db.execute (sqlstring).fetchall ())
    db.close ()
    return result

def main ():
    fit_user_dict = load_fit_user_dict()
    import_to_sunpinyin_user_dict (fit_user_dict)

if __name__ == "__main__":
    main()



#!/usr/bin/python

# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
# 
# Copyright (c) 2007 Sun Microsystems, Inc. All Rights Reserved.
# 
# The contents of this file are subject to the terms of either the GNU Lesser
# General Public License Version 2.1 only ("LGPL") or the Common Development and
# Distribution License ("CDDL")(collectively, the "License"). You may not use this
# file except in compliance with the License. You can obtain a copy of the CDDL at
# http://www.opensource.org/licenses/cddl1.php and a copy of the LGPLv2.1 at
# http://www.opensource.org/licenses/lgpl-license.php. See the License for the 
# specific language governing permissions and limitations under the License. When
# distributing the software, include this License Header Notice in each file and
# include the full text of the License in the License file as well as the
# following notice:
# 
# NOTICE PURSUANT TO SECTION 9 OF THE COMMON DEVELOPMENT AND DISTRIBUTION LICENSE
# (CDDL)
# For Covered Software in this distribution, this License shall be governed by the
# laws of the State of California (excluding conflict-of-law provisions).
# Any litigation relating to this License shall be subject to the jurisdiction of
# the Federal Courts of the Northern District of California and the state courts
# of the State of California, with venue lying in Santa Clara County, California.
# 
# Contributor(s):
# 
# If you wish your version of this file to be governed by only the CDDL or only
# the LGPL Version 2.1, indicate your decision by adding "[Contributor]" elects to
# include this software in this distribution under the [CDDL or LGPL Version 2.1]
# license." If you don't indicate a single choice of license, a recipient has the
# option to distribute your version of this file under either the CDDL or the LGPL
# Version 2.1, or to extend the choice of license to its licensees as provided
# above. However, if you add LGPL Version 2.1 code and therefore, elected the LGPL
# Version 2 license, then the option applies only if the new code is made subject
# to such option by the copyright holder. 

import sys
import getopt
import codecs
import struct

from imdict import IMDict
from trie import match_longest, get_ambiguious_length
from utils import read_ch_sentences

def usage():
    print '''
Usage:
mmseg.py -d dict_file [-f (text|bin)] [-i] [-s STOK_ID] [-a AMBI_ID] corpus_file

  -d --dict:
    The dictionary file (in UTF-8 encoding) to be used.
  -f --format:
    Output format, can be 'text' or 'bin'. Default is 'bin'.
    Normally, in text mode, word text are output, while in binary mode,
    the integer of the word-ids are writed to stdout.
  -i --show-id:
    Show Id info. In text output format, attach id after known words. 
  -s --stok-id:
    Sentence token id. Default 10.
    It will be write to output in binary mode after every sentence.
  -a --ambi-id:
    Ambiguious means ABC => A BC or AB C. If specified (AMBI-ID != 0), 
    The sequence ABC will not be segmented, in binary mode, the AMBI-ID 
    is written out; in text mode, <ambi>ABC</ambi> will be output. Default 
    is 9.
'''

options={'show-id':       False, 
         'format' :       'bin', 
         'stok-id':       10,
         'ambi-id':       9}

def parse_options(args):
    try:
        opts, args = getopt.getopt(args, "hid:f:s:a:", ["help", "show-id", "dict=", "format=", "stok-id=", "ambi-id="])
    except getopt.GetoptError, err:
        print str(err) 
        sys.exit(1)

    for opt,val in opts:
        if opt in ('-h', '--help'):
            usage()
            sys.exit()
        elif opt in ('-d', '--dict'):
            options['dict'] = val
        elif opt in ('-i', '--show-id'):
            options['show-id'] = True
        elif opt in ('-f', '--format'):
            if val in ('bin', 'text'):
                options['format'] = val
        elif opt in ('-s', '--stok-id'):
            options['stok-id'] = int(val)
        elif opt in ('-val', '--ambi-id'):
            options['ambi-id'] = int(val)

    if 'dict' not in options:
        usage()
        sys.exit(1)

    if args:
        options['corpus'] = args[0]

def output_word(wid, word):
    if options['format'] == 'text':
        if wid == options['ambi-id']:
            word = '<ambi>'+word+'</ambi>'
        if options['show-id']:
            word = word+'('+str(wid)+')'
        sys.stdout.write('%s ' % word.encode('UTF-8'))
    else:
        sys.stdout.write(struct.pack('l', wid))

def process_file(file, dict):
    for line in read_ch_sentences(file):
        length = len(line)
        i = 0
        while (i < length):
            strbuf = line[i:]
            wid, l = match_longest(dict, strbuf)
            if wid == 0:
                l = 1
            else:
                ambi_len = get_ambiguious_length(dict, strbuf, l)
                if ambi_len > l:
                    wid, l = options['ambi-id'], ambi_len

            output_word (wid, strbuf[:l])
            i += l

        output_word (options['stok-id'], '\n')

if __name__ == "__main__":
    parse_options(sys.argv[1:])

    dict = IMDict(options['dict'])
    
    try:    file = codecs.open(options['corpus'], "r", "UTF-8")
    except: file = codecs.getreader('UTF-8')(sys.stdin)

    process_file (file, dict)
    file.close()



from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

setup(
    name='sunpinyin-python',
    version='0.1',
    author = "Yong Sun",
    author_email = "Yong.Sun@Sun.COM",
    url = "http://www.opensolaris.org/os/project/input-method",
    description = "A Python binding of the SunPinyin ngram model",
    long_description = "A Python binding of the SunPinyin ngram model",
    license = "CDDL/LGPLv2.1",

    ext_modules=[
        Extension(
            "pyslm", ["pyslm.pyx", 
                      "../src/slm/slm.cpp"], 
            language="c++", 
            include_dirs=["../src/slm", ".."], 
            define_macros=[("HAVE_CONFIG_H", None)]
        ),
        Extension(
            "pytrie", ["pytrie.pyx", 
                       "../src/portability.cpp", 
                       "../src/lexicon/pytrie.cpp", 
                       "../src/lexicon/pinyin_data.cpp"], 
            language="c++", 
            include_dirs=["../src", "../src/lexicon", ".."], 
            define_macros=[("HAVE_CONFIG_H", None)]
        ),
    ],
    cmdclass = {'build_ext': build_ext}
)



#!/usr/bin/python
# -*- coding: UTF-8 -*-

from pyslm import Slm, SlmState
from pytrie import PinyinTrie, PinyinTrieNode, WordInfo

def test_pyslm ():
    slm = Slm ()
    if not slm.load ("../data/lm_sc.t3g"):
        return

    pr, result = slm.transfer (SlmState(0,0), 58614)
    print "pr =", pr, "\tresult = %s" % result
    
    pr, result = slm.transfer (result, 75956)
    print "pr =", pr, "\tresult = %s" % result
    
    pr, result = slm.transfer (result, 84582)
    print "pr =", pr, "\tresult = %s" % result
    
    his = slm.history_state_of (result)
    print "his = %s" % his
    
    slm.historify (result)
    print "result = %s" % result

    print 'last_word_id =', slm.last_word_id (result)

    slm.free ()

def test_pytrie ():
    trie = PinyinTrie()
    if not trie.load ("../data/pydict_sc.bin"):
        return

    root = trie.get_root_node ()
    node = trie.transfer (root, 0x1000)
    for w in node.get_words ():
        print w

    print trie.is_valid (node, False, 0)
    print trie[10000]

    print trie.get_symbol_id (u'。')
    trie.free ()

test_pyslm()
test_pytrie()



#!/usr/bin/python
import os, sys
import struct
import sqlite3 as sqlite
from pinyin_data import valid_syllables, decode_syllable, initials, finals

def get_userdict_path ():
    homedir = os.environ.get("HOME")

    if sys.platform == "darwin":
        return homedir+"/Library/Application Support/SunPinyin/userdict"

    # FIXME: not sure how to get the ibus version or wrapper type (xim or ibus)
    if os.path.exists (homedir+"/.cache/ibus/sunpinyin"):
        return homedir+"/.cache/ibus/sunpinyin/userdict"
        
    if os.path.exists (homedir+"/.ibus/sunpinyin"):
        return homedir+"/.ibus/sunpinyin/userdict"
    
    if os.path.exists (homedir+"/.sunpinyin"):
        return homedir+"/.sunpinyin/userdict"

    raise "Can not detect sunpinyin's userdict!"

def get_sysdict_path ():
    if sys.platform == "darwin":
        homedir = os.environ.get("HOME")
        sysdict_path = "/Library/Input Methods/SunPinyin.app/Contents/Resources/pydict_sc.bin"
        if os.path.exists (homedir + sysdict_path):
            return homedir + sysdict_path
        else:
            return sysdict_path

    return "/usr/lib/sunpinyin/data/pydict_sc.bin"

def load_system_dict ():
    sysdict_path = get_sysdict_path ()
    f = open (sysdict_path, "rb")
    
    f.seek(8)
    word_offset = struct.unpack ('I', f.read(4))[0]
    f.seek (word_offset)

    words = set()
    str = f.read()
    
    for w in str.decode('UTF-32').split('\0'):
        if w:
            words.add (w)
    
    f.close()
    return words

def import_to_sunpinyin_user_dict (records, userdict_path=''):
    userdict_path = userdict_path if userdict_path else get_userdict_path()
    db = sqlite.connect (userdict_path)

    sysdict = load_system_dict()

    sqlstring = """
            CREATE TABLE IF NOT EXISTS dict(
            id INTEGER PRIMARY KEY, len INTEGER,
            i0 INTEGER, i1 INTEGER, i2 INTEGER, i3 INTEGER, i4 INTEGER, i5 INTEGER,
            f0 INTEGER, f1 INTEGER, f2 INTEGER, f3 INTEGER, f4 INTEGER, f5 INTEGER,
            utf8str TEXT, UNIQUE (utf8str));
            """
    db.executescript (sqlstring)

    batch_count = 0

    for (pystr, utf8str) in records:
        try:
            syllables = [valid_syllables[s] for s in pystr.split("'")]
        except:
            print "[%s] has un-recognized syllables, ignoring this record!" % pystr
            continue

        if len (syllables) < 2 or len (syllables) > 6:
            print "[%s] is too long or too short for sunpinyin userdict" % utf8str
            continue

        if utf8str in sysdict:
            #print "[%s] is already in sunpinyin's sysdict" % utf8str
            continue

        record = [0]*14
        record[0] = len (syllables)
        record[13] = utf8str

        c = 1
        for s in syllables:
            i, f = s>>12, (s&0x00ff0)>>4
            if i and not f:
                break; 
            record[c] = i
            record[c+1] = f
            c += 2
        else:
            sqlstring = """
                    INSERT INTO dict (len, i0, f0, i1, f1, i2, f2, i3, f3, i4, f4, i5, f5, utf8str)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
                    """
            try:
                db.execute (sqlstring, record)
                #print "[%s] is imported into sunpinyin's userdict" % utf8str

                batch_count += 1
                if batch_count == 100:
                    db.commit ()
                    batch_count = 0

            except:
                #print "[%s] is already in sunpinyin's userdict" % utf8str
                pass

    db.commit()
    db.close()

def export_sunpinyin_user_dict (userdict_path=''):
    userdict_path = userdict_path if userdict_path else get_userdict_path()
    db = sqlite.connect (userdict_path)

    sqlstring = "SELECT * FROM dict"
    result = list (db.execute (sqlstring).fetchall ())

    for record in result:
        id   = record[0]
        l    = record[1]
        i    = record[2:8]
        f    = record[8:14]
        str  = record[-1]
        syls = [initials[i[x]] + finals[f[x]] for x in range(l)]
        print str.encode ('UTF-8'), id, "'".join(syls) 
        
if __name__ == "__main__":
    export_sunpinyin_user_dict ()



#!/usr/bin/python

import os, sys
import codecs
from importer import import_to_sunpinyin_user_dict

def load_fcitx_user_dict (fname):
    result = []
    f = codecs.open (fname, "r", "GB18030")
    for l in f:
        if l[0] in ('\n'):
            continue

        pystr, utf8str = l.strip().split(" ")
        result.append ((pystr, utf8str))

    return result

def main ():
    if len (sys.argv) != 2:
        os.system ("/usr/bin/mb2org ~/.fcitx/pyusrphrase.mb > /tmp/fcitx_userdict_gb.txt")
        fcitx_user_dict = load_fcitx_user_dict("/tmp/fcitx_userdict_gb.txt")
        os.system ("rm -rf /tmp/fcitx_userdict_gb.txt")
    else:
        fcitx_user_dict = load_fcitx_user_dict(sys.argv[1])

    import_to_sunpinyin_user_dict (fcitx_user_dict)

if __name__ == "__main__":
    main()



#!/usr/bin/python

import os, sys
import codecs
from importer import import_to_sunpinyin_user_dict

def load_sogou_user_dict (fname):
    result = []
    f = codecs.open (fname, "r", "UTF-16")

    for l in f:
        if l[0] in (';', '\n', '\r'):
            continue

        pystr, utf8str = l.strip().split(" ")
        result.append ((pystr[1:], utf8str))

    return result

def main ():
    if len (sys.argv) != 2:
        print "Please specify the Sogou PinYin exported user dict file!"
        exit (1)

    sogou_user_dict = load_sogou_user_dict(sys.argv[1])
    import_to_sunpinyin_user_dict (sogou_user_dict)

if __name__ == "__main__":
    main()
